# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from erniebot_agent.agents.callback.handlers.base import CallbackHandler
from erniebot_agent.agents.callback.handlers.logging_handler import LoggingHandler


def get_default_callbacks() -> List[CallbackHandler]:
    """Get default callback handlers."""
    return [LoggingHandler()]


def get_no_ellipsis_callback() -> List[CallbackHandler]:
    return [LoggingHandler()]
